#####################################################
## Class to store an ngram ME model
#####################################################
import sys
from math import exp
import random
import pickle
import json

import scipy
import sklearn
from numpy import ndarray
from sklearn.linear_model import LogisticRegression

## ---------------------------------
## Auxiliar: iterate over ngrams in given stream f
## ---------------------------------
def ngrams(f,n) :
    ng = "._"
    ch = f.read(1).lower()
    if ch.isspace() : ch = '_'
    ng += ch
    while ch :
        yield ng
        ch = f.read(1).lower()
        if ch.isspace() : ch = '_'
        ng = (ng+ch)[1:]
       
## ---------------------------------
## Auxiliar: convert a bigram into a feature vector
## ---------------------------------
def extract_features(xy) :
   return ["1st:"+xy[0], "2nd:"+xy[1], "bigr:"+xy[0:2] ]     

###############################################
## Class Dataset:
##    - Loads a text and stores it as a sequence of trigram examples
##    - Each trigram is a context (feature vector with information about
##      the conditioning bigram) plus a gold label (actual 3rd character)
##    - Numbers the features and creates indexes to retrive feature name
##      from index and viceversa
##    - Provides access to dataset as a CSR matrix
###############################################

class Dataset :
    # ----- read text, extract trigrams, convert each trigram to a feature vector.
    # ----- Store vectors and create feature index
    def __init__(self, datafile) :
        self.examples = []
        self.index = {"feat2idx":{}, "idx2feat":{}}
        nf = 0
        print("Loading data")
        with open(datafile) as df :
            for ng in ngrams(df, 3) :
               features = extract_features(ng[:2])
               label = ng[2]  
               self.examples.append({"label": label, "features": features})
               for f in features :
                  if f not in self.index["feat2idx"] :
                     i = len(self.index["feat2idx"])
                     self.index["feat2idx"][f] = i
                     self.index["idx2feat"][i] = f
        
    ## ------ allow access to feature indexes
    def get_feature_index(self) :
        return self.index

    ## ------ return dataset as a sparse matrix, plus associated gold labels
    def csr_matrix(self) :
        rowi = [] # row (example number)
        colj = [] # column (feature number)
        data = [] # value (1 or 0 since we use binary features)
        Y = [] # ground truth
        nex = 0 # example  counter (each ngram is one example)

        # for each example
        for ex in self.examples :
            Y.append(ex["label"])
            for f in ex["features"] :
                data.append(1)
                rowi.append(nex)
                colj.append(self.index["feat2idx"][f]) 
            # next example
            nex += 1
            
        X = scipy.sparse.csr_matrix((data, (rowi, colj)))  
        return X,Y

###############################################
##  class MEM:
##    - Creates a Maximum entropy model for trigrams
##    - Allows training new models and saving to disk
##    - Allows loading saved models
##    - Allows querying the model about probabilities
#       · Probability of a 3rd character following given bigram
#       · Probability distribution of all characters after given bigram
#       · Generate a random character after given bigram, according to distribution
###############################################

class MEM:

    ## --------------------------------------------------
    ## Constructor: Load model from file
    ## --------------------------------------------------
    def __init__(self, modelfile=None):
        if modelfile is not None:
            # modelfile given, assume it is an existing model and load it       
            self.load(modelfile)
                
        else :  # no modelfile given, create new empty model
            self.model = {}
                
    ## --------------------------------------------------
    ## train a model on given data with given params
    ## --------------------------------------------------
    def train(self, datafile, params=None):
        if params is None: params = {}        
        # extract parameters if provided. Use default if empty
        C = float(params['C']) if 'C' in params else 1.0
        solver = params['solver'] if 'solver' in params else 'lbfgs'
        maxit = params['max_iter'] if 'max_iter' in params else 1500
        jobs = params['n_jobs'] if 'n_jobs' in params else 8

        # create and train empty classifier with given parameters
        self.model["mem"] = LogisticRegression(verbose=1,
                                               fit_intercept=False,
                                               C=C,
                                               solver=solver,
                                               max_iter=maxit,
                                               n_jobs=jobs)    
        # load dataset
        ds = Dataset(datafile)
        self.model["index"] = ds.get_feature_index()

        # Get training instances as a sparse matrix
        print("Encoding as CSR matrix")
        X,Y = ds.csr_matrix()

        # train classifier
        print("Training")
        self.model["mem"].fit(X,Y)

    
    ## --------------------------------------------------
    ## load saved model
    ## --------------------------------------------------
    def load(self, fmodelname) :
        with open(fmodelname, 'rb') as df: 
           self.model = pickle.load(df)

    ## --------------------------------------------------
    ## save model
    ## --------------------------------------------------
    def save(self, fmodelname) :
        pickle.dump(self.model, open(fmodelname+".pck", 'wb'))
 
    ## --------------------------------------------------
    ## compute probability distribution for all possible classes given features in 'feats'
    ## --------------------------------------------------
    def prob_dist_z(self,feats) :

        classes = self.model["mem"].classes_
        lambdas = self.model["mem"].coef_
        index = self.model["index"]["feat2idx"]

        p = {}
        z = 0        
        for c in range(len(classes)) :
            ## add lambda_i * f_i for each example feature seen in training
            s=0 
            for f in feats :
                if f in index : 
                   fi = index[f]
                   s = s + lambdas[c,fi]
            
            label = classes[c]
            p[label] = exp(s)
            z = z + p[label]

        # normalize class probabilities
        for c in classes :
            p[c] = p[c]/z

        return p

    ## --------------------------------------------------
    ## compute P(clas|context)
    ## --------------------------------------------------
    def conditional_prob(self, context, clas) :
        dist = self.prob_dist_z(context)
        return dist[clas]
    
    
    ## ---------------------------------
    # return a random prediction, consistent with model distribution     
    ## ---------------------------------
    def random_transition(self, context) :

       # obtain distribution of predictions given current features
       dist = self.prob_dist_z(context)
       # randomly select a prediction according to distribution
       r = random.random()
       s = 0.0
       for c in dist :
           s = s + dist[c]
           if (s>=r) : return c



#######################################    
## MAIN ---------------------
## 
## -- read text in ngrams, train model, 
## -- save MEM model

if __name__ == "__main__" :

    ## -- create empty model
    model = MEM()
    # train it with given data
    model.train(sys.argv[1])
    # save to given name
    model.save(sys.argv[2])


